﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System.IO;
using UnityEditor;

namespace UPP.Engine
{
    public static class ClassManager
    {
        private static ClassesDB _db = null;
        public static ClassesDB DB { get {
                if (_db == null)
                    LoadDatabase();
                return _db;
            } }

        private static void CreateDatabase()
        {
            _db = ScriptableObject.CreateInstance<ClassesDB>();
            AssetDatabase.CreateAsset(_db, Path.Get(Path.KEY_OUTPUT_CACHE_CLASSES_DB));

            AssetDatabase.Refresh();
            EditorUtility.SetDirty(_db);
            AssetDatabase.SaveAssets();

            _db.database = new List<ClassDBRecord>();
        }

        private static void LoadDatabase()
        {
            _db = AssetDatabase.LoadAssetAtPath<ClassesDB>(Path.Get(Path.KEY_OUTPUT_CACHE_CLASSES_DB));
            if (_db == null)
                CreateDatabase();
        }

        private static string _lastManipulatedClassName = "";
        public static string LastManipulatedClassName { get { return _lastManipulatedClassName; } }

        private static string OutputCachePath
        {
            get
            {
                string s = Settings.Get(Path.Resolve(Path.KEY_OUTPUT_CACHE_CLASSES_PATH));

                if (!Directory.Exists(s))
                    Directory.CreateDirectory(s);

                return s;
            }
        }

        private static string NameToPath(string name, bool virtualTokens)
        {
            if (!virtualTokens)
                return OutputCachePath + "/" + name + ".asset";
            else
                return Path.Token(Path.KEY_OUTPUT_CACHE_CLASSES_PATH) + "/" + name + ".asset";
        }

        public static ClassRecord GetClassByMachineName(string name)
        {
            ClassRecord rec = AssetDatabase.LoadAssetAtPath<ClassRecord>(NameToPath(name, false));
            return rec;
        }

        public static bool ClassExists(string name)
        {
            ClassRecord rec = GetClassByMachineName(name);
            return (rec != null);
        }

        public static List<ClassRecord> FindChildren(ClassRecord parent)
        {
            List<ClassRecord> list = new List<ClassRecord>();

            foreach (var r in DB.database)
            {
                var c = GetClassByMachineName(r.machineName);
                if (c.machineParent == parent.machineName)
                {
                    list.Add(c);
                }
            }
            return list;
        }

        public static bool Remove(string machineName, bool confirmationDialog = true, bool fixChildren =true)
        {
            ClassRecord rec = GetClassByMachineName(machineName);
            if (rec == null)
                return false;

            List<ClassRecord> children = new List<ClassRecord>();
            bool childrenAlreadyDone = false;

            if (confirmationDialog)
            {
                if (EditorUtility.DisplayDialog(
                    "Remove class " + rec.className + "?",
                    "Do you wish to remove the class " + rec.className + " (" + rec.machineName + ")?",
                    "Remove", "Cancel"
                    ))
                {
                    if (fixChildren)
                    {
                        var ch = FindChildren(rec);
                        if (ch.Count > 0)
                        {
                            if (EditorUtility.DisplayDialog(
                                "Are you sure?",
                                "There are still some classes (count: " + ch.Count + ") that are children of this class. Their parent will be set to a parent of the removed class. Still proceed?",
                                "Proceed", "Cancel"
                                ))
                            {
                                children = ch;
                                childrenAlreadyDone = true;
                            }
                            else return false;
                        }
                    }
                }
                else return false;
            }

            string parentOfParent = rec.machineParent;

            if (!childrenAlreadyDone)
                children = FindChildren(rec);

            AssetDatabase.Refresh();
            ClassDBRecord dbrec = DB.database.Find(x => x.machineName == rec.machineName);
            DB.database.Remove(dbrec);
            AssetDatabase.DeleteAsset(NameToPath(rec.machineName, false));
            ScriptableObject.DestroyImmediate(rec);
            EditorUtility.SetDirty(DB);
            foreach (var c in children)
            {
                c.machineParent = parentOfParent;
                EditorUtility.SetDirty(c);
            }



            AssetDatabase.SaveAssets();
            _lastManipulatedClassName = machineName;
            return true;
        }

        public static void ForceSave(string classMachineName)
        {
            AssetDatabase.Refresh();
            EditorUtility.SetDirty(GetClassByMachineName(classMachineName));
            EditorUtility.SetDirty(DB);
            AssetDatabase.SaveAssets();
        }

        public static bool Create(string name, string machineName, string parent = "", string description = "", bool isAbstract = true, bool checkForExistence = true)
        {
            if (machineName == "" || (checkForExistence && ClassExists(machineName)))
            {
                EditorUtility.DisplayDialog("Can't create the class", "Class with the name \"" + machineName + "\" already exists and can't be created again, or the name is invalid.", "OK");
                return false;
            }

            ClassRecord newClass = ScriptableObject.CreateInstance<ClassRecord>();
            newClass.className = name;
            newClass.machineName = machineName;
            newClass.machineParent = parent;
            newClass.description = description;
            newClass.isAbstract = isAbstract;
            string pathPhys = NameToPath(machineName, false);
            AssetDatabase.CreateAsset(newClass, pathPhys);

            DB.database.Add(new ClassDBRecord(name, machineName));

            AssetDatabase.Refresh();
            EditorUtility.SetDirty(newClass);
            EditorUtility.SetDirty(DB);
            AssetDatabase.SaveAssets();
            _lastManipulatedClassName = machineName;
            return true;
        }

        public static bool Edit(ClassRecord rec, string name, string machineName, string parent = "", string description = "", bool isAbstract = true)
        {
            var byName = GetClassByMachineName(machineName);
            if (machineName == "" || (byName != null && byName != rec))
            {
                EditorUtility.DisplayDialog("Can't edit the class", "Class with the name \"" + machineName + "\" already exists, or the name is invalid.", "OK");
                return false;
            }

            string oldName = rec.machineName;

            if (oldName != machineName)
            {
                ClassDBRecord dbrec = DB.database.Find(x => x.machineName == oldName);
                DB.database.Remove(dbrec);

                ClassRecord newClass = ScriptableObject.CreateInstance<ClassRecord>();
                newClass.className = name;
                newClass.machineName = machineName;
                newClass.machineParent = parent;
                newClass.description = description;
                newClass.isAbstract = isAbstract;
                newClass.diagramX = rec.diagramX;
                newClass.diagramY = rec.diagramY;
                string pathPhys = NameToPath(machineName, false);

                AssetDatabase.DeleteAsset(NameToPath(rec.machineName, false));
                ScriptableObject.DestroyImmediate(rec);

                AssetDatabase.Refresh();

                AssetDatabase.CreateAsset(newClass, pathPhys);
                DB.database.Add(new ClassDBRecord(name, machineName));

                foreach (var dbr in DB.database)
                {
                    var c = ClassManager.GetClassByMachineName(dbr.machineName);
                    if (c.machineParent == oldName)
                    {
                        EditorUtility.SetDirty(c);
                        c.machineParent = machineName;
                    }
                }

                EditorUtility.SetDirty(newClass);

                PropertyManager.RenameClassRutine(oldName, machineName);
            }
            else
            {
                ClassDBRecord dbrec = DB.database.Find(x => x.machineName == oldName);
                dbrec.className = name;

                rec.className = name;
                rec.machineName = machineName;
                rec.machineParent = parent;
                rec.description = description;
                rec.isAbstract = isAbstract;
                EditorUtility.SetDirty(rec);
            }

            EditorUtility.SetDirty(DB);
            AssetDatabase.SaveAssets();
            _lastManipulatedClassName = machineName;
            return true;
        }

    }
}