﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Xml.Serialization;
using UnityEngine;
using UPP.OGUI;

namespace UPP.Engine
{
    [Serializable]
    [PropertyInfo(
        Name = "Boolean",
        MachineName = "UPP_bool",
        Description = "A logical value that can be either True or False."
        )]
    public class PropertyBool : Property<bool>
    {
        private Dictionary<int, bool> data;

        [XmlIgnore]
        public bool DefaultValue = false;

        OGUI.CheckBox defBox;
        OGUI.CheckBox valBox;

        private bool Validated(bool value)
        {
            return value;
        }

        private void ValidateAll()
        {
            foreach (var key in data.Keys)
            {
                data[key] = Validated(data[key]);
            }
        }

        public override void Unpack(string[] data)
        {
            this.data = DictionaryToArray<bool>.Unpack(data, Convert.ToBoolean);
        }

        public override string[] Pack()
        {
            return DictionaryToArray<bool>.Pack(data);
        }

        public override void Set(bool data, int index = 0)
        {
            if (index < 0 || index >= this.ValueCount)
            {
                if (this.ValueCount != 0)
                {
                    // Specific Value count => error; out of bounds.
                    throw new Exception("Property " + this.MachineName + " has only " +
                        this.ValueCount + " values, but index [" + index + "] was used in Set.");
                }
            }

            this.data[index] = Validated(data);
        }

        public override bool Get(int index = 0)
        {
            if (index < 0 || index >= this.ValueCount)
            {
                if (this.ValueCount != 0)
                {
                    throw new Exception("Property " + this.MachineName + " has only " +
                        this.ValueCount + " values, but index [" + index + "] was used in Get.");
                }
            }

            if (this.data.ContainsKey(index))
                return this.data[index];
            else
                return GetDefault();
        }

        protected override string Serialize()
        {
            string dataArrStr = DictionaryToArray<bool>.BasicSerializer.Serialize(data);

            return Serializer.EncapsulateStrings(
                dataArrStr,
                DefaultValue.ToString()
                );
        }

        protected override void Deserialize(string str)
        {
            string dataArrStr;

            string[] s = Serializer.DecapsulateString(str);
            dataArrStr = s[0];
            DefaultValue = Convert.ToBoolean(s[1]);

            data = DictionaryToArray<bool>.BasicSerializer.Deserialize(dataArrStr, Convert.ToBoolean);
        }

        public override string GetDisplayValue(bool shorten)
        {
            int to = 0;
            string s = "";
            string e = "";
            if (!shorten || (ValueCount <= 10 && ValueCount != 0))
                to = this.ValueCount;
            else
            {
                e = "...";
                to = 10;
            }
            for (int i = 0; i < to; i++)
            {
                if (s != "")
                    s += ", ";
                s += Get(i);
            }
            return s + e;
        }

        public override void GUIForSetting(Panel parent)
        {
            Label l1 = new Label("Dafault Value:");
            l1.Parent = parent;
            l1.Position = new Position(10, 10);

            defBox = new CheckBox(DefaultValue, "True");
            defBox.Parent = parent;
            defBox.Position = new Position(100, 10);
            defBox.Size.Width = 100;

        }

        public override void SaveGUIForSetting()
        {
            DefaultValue = defBox.Checked;
        }

        public override bool GetDefault()
        {
            return DefaultValue;
        }

        int ValuePage = 0;

        public override void GUIForValue(Panel parent)
        {
            Label l1 = new Label("Value:");
            l1.Parent = parent;
            l1.Position = new Position(10, 10);

            valBox = new CheckBox(Get(0));
            valBox.Parent = parent;
            valBox.Position = new Position(100, 10);
            valBox.Size.Width = parent.Size.Width - 110;
            valBox.Anchor = new Anchor(true, true, false, true);
            valBox.OnChange += ValBox_OnChange;
        }

        private void ValBox_OnChange(object sender)
        {
            Set(valBox.Checked, ValuePage);
        }

        public override void GUIForValueSetPage(int index)
        {
            Set(valBox.Checked, ValuePage);
            ValuePage = index;
            valBox.Checked = Get(index);
        }

        public override string GenerateSourceAttributes()
        {
            return "";
        }

        public override string GenerateSourceDeclaration(bool defaultValue)
        {
            string n = (defaultValue) ? "" : "new ";
            string mod = ExposeToInstances ? "public" : "public const";
            return n + mod + " bool " + MachineName + " = " +
                ((defaultValue) ? GetDefault().ToString() : Get().ToString()) + ";";
        }
    }

}