﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Xml.Serialization;
using UnityEngine;
using UPP.OGUI;

namespace UPP.Engine
{
    [Serializable]
    [PropertyInfo(
        Name = "Real Number",
        MachineName = "UPP_double",
        Description = "A real number with a floaing point. Both positive and negative numbers are allowed, unless limited through Minimum and/or Maximum."
        )]
    public class PropertyDouble : Property<double>
    {
        private Dictionary<int, double> data;

        [XmlIgnore]
        public double Maximum = double.MaxValue;
        [XmlIgnore]
        public double Minimum = double.MinValue;
        [XmlIgnore]
        public double DefaultValue = 0;

        OGUI.TextBox maxBox;
        OGUI.TextBox minBox;
        OGUI.TextBox defBox;
        OGUI.CheckBox useMin;
        OGUI.CheckBox useMax;

        OGUI.TextBox valBox;

        private double Validated(double value)
        {
            return Math.Max(Math.Min(value, Maximum), Minimum);
        }

        private void ValidateAll()
        {
            foreach (var key in data.Keys)
            {
                data[key] = Validated(data[key]);
            }
        }

        public override void Unpack(string[] data)
        {
            this.data = DictionaryToArray<double>.Unpack(data, Convert.ToDouble);
        }

        public override string[] Pack()
        {
            return DictionaryToArray<double>.Pack(data);
        }

        public override void Set(double data, int index = 0)
        {
            if (index < 0 || index >= this.ValueCount)
            {
                if (this.ValueCount != 0)
                {
                    // Specific Value count => error; out of bounds.
                    throw new Exception("Property " + this.MachineName + " has only " +
                        this.ValueCount + " values, but index [" + index + "] was used in Set.");
                }
            }

            this.data[index] = Validated(data);
        }

        public override double Get(int index = 0)
        {
            if (index < 0 || index >= this.ValueCount)
            {
                if (this.ValueCount != 0)
                {
                    throw new Exception("Property " + this.MachineName + " has only " +
                        this.ValueCount + " values, but index [" + index + "] was used in Get.");
                }
            }

            if (this.data.ContainsKey(index))
                return this.data[index];
            else
                return GetDefault();
        }

        protected override string Serialize()
        {
            string dataArrStr = DictionaryToArray<double>.BasicSerializer.Serialize(data);

            return Serializer.EncapsulateStrings(
                dataArrStr,
                Minimum.ToString(),
                Maximum.ToString(),
                DefaultValue.ToString()
                );
        }

        protected override void Deserialize(string str)
        {
            string dataArrStr;

            string[] s = Serializer.DecapsulateString(str);
            dataArrStr = s[0];
            Minimum = Convert.ToDouble(s[1]);
            Maximum = Convert.ToDouble(s[2]);
            DefaultValue = Convert.ToDouble(s[3]);

            data = DictionaryToArray<double>.BasicSerializer.Deserialize(dataArrStr, Convert.ToDouble);
        }

        public override string GetDisplayValue(bool shorten)
        {
            int to = 0;
            string s = "";
            string e = "";
            if (!shorten || (ValueCount <= 10 && ValueCount != 0))
                to = this.ValueCount;
            else
            {
                e = "...";
                to = 10;
            }
            for (int i = 0; i < to; i++)
            {
                if (s != "")
                    s += ", ";
                s += Get(i);
            }
            return s + e;
        }

        public override void GUIForSetting(Panel parent)
        {
            Label l1 = new Label("Minimum:");
            l1.Parent = parent;
            l1.Position = new Position(10, 10);

            Label l2 = new Label("Maximum:");
            l2.Parent = parent;
            l2.Position = new Position(10, 35);

            Label l3 = new Label("Default value:");
            l3.Parent = parent;
            l3.Position = new Position(10, 60);

            minBox = new TextBox(Minimum.ToString());
            minBox.Parent = parent;
            minBox.Position = new Position(100, 10);
            minBox.Size.Width = 100;
            minBox.OnChange += Box_OnChange;

            maxBox = new TextBox(Maximum.ToString());
            maxBox.Parent = parent;
            maxBox.Position = new Position(100, 35);
            maxBox.Size.Width = 100;
            maxBox.OnChange += Box_OnChange;

            defBox = new TextBox(DefaultValue.ToString());
            defBox.Parent = parent;
            defBox.Position = new Position(100, 60);
            defBox.Size.Width = 100;

            useMin = new CheckBox(Minimum != double.MinValue, "Limit");
            useMin.Parent = parent;
            useMin.Position = new Position(205, 10);
            useMin.Size.Width = 105;
            useMin.OnChange += Use_OnChange;

            useMax = new CheckBox(Maximum != double.MaxValue, "Limit");
            useMax.Parent = parent;
            useMax.Position = new Position(205, 35);
            useMax.Size.Width = 105;
            useMax.OnChange += Use_OnChange;

            maxBox.Enabled = useMax.Checked;
            minBox.Enabled = useMin.Checked;
        }

        private void Box_OnChange(object sender)
        {
            ClampDefValue();
        }

        private void defBoxCorrect()
        {
            var min = Convert.ToDouble(minBox.Value);
            var max = Convert.ToDouble(maxBox.Value);
            var def = Convert.ToDouble(defBox.Value);
            defBox.Value = Math.Max(Math.Min(def, max), min).ToString();
        }

        private void ClampDefValue()
        {
            defBoxCorrect();
        }

        private void Use_OnChange(object sender)
        {
            maxBox.Enabled = useMax.Checked;
            minBox.Enabled = useMin.Checked;
            ClampDefValue();
        }

        public override void SaveGUIForSetting()
        {
            Maximum = (useMax.Checked) ? Convert.ToDouble(maxBox.Value) : double.MaxValue;
            Minimum = (useMin.Checked) ? Convert.ToDouble(minBox.Value) : double.MinValue;
            DefaultValue = Convert.ToDouble(defBox.Value);
        }

        public override double GetDefault()
        {
            return DefaultValue;
        }

        int ValuePage = 0;

        public override void GUIForValue(Panel parent)
        {
            Label l1 = new Label("Value:");
            l1.Parent = parent;
            l1.Position = new Position(10, 10);

            valBox = new TextBox(Get(0).ToString());
            valBox.Parent = parent;
            valBox.Position = new Position(100, 10);
            valBox.Size.Width = parent.Size.Width - 110;
            valBox.Anchor = new Anchor(true, true, false, true);
            valBox.OnChange += ValBox_OnChange;
        }

        private void ValBox_OnChange(object sender)
        {
            Set(Convert.ToDouble(valBox.Value), ValuePage);
        }

        public override void GUIForValueSetPage(int index)
        {
            Set(Convert.ToDouble(valBox.Value), ValuePage);
            ValuePage = index;
            valBox.Value = Get(index).ToString();
        }

        public override string GenerateSourceAttributes()
        {
            if (Minimum != double.MinValue || Maximum != double.MaxValue)
                return "[Range(" + Minimum.ToString() + ", " + Maximum.ToString() + ")]";
            else
                return "";
        }

        public override string GenerateSourceDeclaration(bool defaultValue)
        {
            string n = (defaultValue) ? "" : "new ";
            string mod = ExposeToInstances ? "public" : "public const";
            return n + mod + " int " + MachineName + " = " +
                ((defaultValue) ? GetDefault().ToString() : Get().ToString()) + ";";
        }
    }

}