﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using System.Xml.Serialization;
using UnityEngine;
using UPP.OGUI;

namespace UPP.Engine
{
    [PropertyInfo(
        Name = "Text",
        MachineName = "UPP_string",
        Description = "Basic text property with a specified maximum length.")]
    public class PropertyString : Property<string>
    {
        private Dictionary<int, string> data;

        [XmlIgnore]
        public int MaxLength = 255;

        NumBox nbox;
        TextBox valBox;

        public override string GetDisplayValue(bool shorten)
        {
            int to = this.ValueCount;
            if (to == 0)
                to = data.Keys.Max() + 1;
            string s = "";
            for (int i = 0; i < to; i++)
            {
                if (s != "")
                    s += ", ";
                string x = Get(i);
                x = x.Replace("\n", " ");
                x = x.Replace("\r", " ");
                x = x.Replace("\t", " ");
                x = Regex.Replace(x, @"\s+", " ");
                s += x;
                if (s.Length >= 255)
                    break;
            }

            if (shorten)
            {
                
                if (s.Length > 255)
                    s = s.Substring(0, 255);
            }
            return s;
        }

        protected override string Serialize()
        {
            string dataArrStr = DictionaryToArray<string>.BasicSerializer.Serialize(data);

            return Serializer.EncapsulateStrings(
                dataArrStr, 
                MaxLength.ToString()
            );
        }

        private string ToStr(object o)
        {
            return o.ToString();
        }

        protected override void Deserialize(string str)
        {
            string dataArr;

            string[] s = Serializer.DecapsulateString(str);
            dataArr = s[0];
            MaxLength = Convert.ToInt32(s[1]);

            data = DictionaryToArray<string>.BasicSerializer.Deserialize(dataArr, ToStr);
        }

        private string Validated(string value)
        {
            return value.Substring(0, Math.Min(MaxLength, value.Length));
        }

        private void ValidateAll()
        {
            foreach (var key in data.Keys)
            {
                data[key] = Validated(data[key]);
            }
        }

        public override void Unpack(string[] data)
        {
            this.data = DictionaryToArray<string>.Unpack(data, ToStr);
        }

        public override string[] Pack()
        {
            return DictionaryToArray<string>.Pack(data);
        }

        public override void Set(string data, int index = 0)
        {
            if (index < 0 || index >= this.ValueCount)
            {
                if (this.ValueCount != 0)
                {
                    // Specific Value count => error; out of bounds.
                    throw new Exception("Property " + this.MachineName + " has only " +
                        this.ValueCount + " values, but index [" + index + "] was used in Set.");
                }
            }

            this.data[index] = Validated(data);
        }

        public override string Get(int index = 0)
        {
            if (data == null)
                return "";

            if (index < 0 || index >= this.ValueCount)
            {
                if (this.ValueCount != 0)
                {
                    throw new Exception("Property " + this.MachineName + " has only " +
                        this.ValueCount + " values, but index [" + index + "] was used in Get.");
                }
            }

            if (this.data.ContainsKey(index))
                return this.data[index];
            else
                return GetDefault();
        }

        public override void GUIForSetting(OGUI.Panel parent)
        {
            OGUI.Label l = new OGUI.Label("Maximum length:");
            l.Parent = parent;
            l.Position = new Position(10, 10);

            nbox = new OGUI.NumBox(MaxLength, 0, 100000);
            nbox.Parent = parent;
            nbox.Anchor = new Anchor(true, true, false, true);
            nbox.Position = new Position(140, 10);
            nbox.Size.Width = parent.Size.Width - 150;
        }

        public override void SaveGUIForSetting()
        {
            MaxLength = nbox.Value;
        }

        public override string GetDefault()
        {
            return "";
        }

        int ValuePage = 0;

        public override void GUIForValue(Panel parent)
        {
            Label l1 = new Label("Value:");
            l1.Parent = parent;
            l1.Position = new Position(10, 10);

            valBox = new TextBox(Get(0), MaxLength);
            valBox.Parent = parent;
            valBox.Position = new Position(100, 10);
            valBox.Size.Width = parent.Size.Width - 110;
            valBox.Anchor = new Anchor(true, true, false, true);
            valBox.OnChange += ValBox_OnChange;
        }

        private void ValBox_OnChange(object sender)
        {
            Set(valBox.Value, ValuePage);
        }

        public override void GUIForValueSetPage(int index)
        {
            ValuePage = index;
            valBox.Value = Get(index);
        }

        public override string GenerateSourceAttributes()
        {
            return "";
        }

        public override string GenerateSourceDeclaration(bool defaultValue)
        {
            string n = (defaultValue) ? "" : "new ";
            string mod = ExposeToInstances ? "public" : "public const";
            return n + mod + " string " + MachineName + " = \"" +
                ((defaultValue) ? GetDefault() : Get()) + "\";";
        }
    }
}