﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using UnityEditor;
using UnityEngine;

namespace UPP.Engine
{

    public static class PropertyManager
    {
        public class ExposeFilter
        {
            public enum Filter
            {
                No = -1,
                DontCare = 0,
                Yes = 0
            }

            public Filter ExposeToTemplates { get; set; }
            public Filter ExposeToInstances { get; set; }

            public ExposeFilter(Filter exposeToTemplates = Filter.DontCare, Filter exposeToInstances = Filter.DontCare)
            {
                ExposeToTemplates = ExposeToTemplates;
                exposeToInstances = ExposeToInstances;
            }

            public bool MatchesFilter(bool exposeToTemplates, bool exposeToInstances)
            {
                return (ExposeToTemplates == Filter.DontCare ||
                    (ExposeToTemplates == Filter.Yes && exposeToTemplates) ||
                    (ExposeToTemplates == Filter.No && !exposeToTemplates))
                    &&
                    (ExposeToInstances == Filter.DontCare ||
                    (ExposeToInstances == Filter.Yes && exposeToInstances) ||
                    (ExposeToInstances == Filter.No && !exposeToInstances));
            }
        }


        private static string _lastManipulatedClassName = "";
        public static string LastManipulatedClassName { get { return _lastManipulatedClassName; } }
        private static string _lastManipulatedPropertyName = "";
        public static string LastManipulatedPropertyName { get { return _lastManipulatedPropertyName; } }

        private static PropertiesDB _db = null;
        public static PropertiesDB DB
        {
            get
            {
                if (_db == null)
                    LoadDatabase();
                return _db;
            }
        }

        private static void CreateDatabase()
        {
            _db = ScriptableObject.CreateInstance<PropertiesDB>();
            AssetDatabase.CreateAsset(_db, Path.Get(Path.KEY_OUTPUT_CACHE_PROPERTIES_DB));

            AssetDatabase.Refresh();
            EditorUtility.SetDirty(_db);
            AssetDatabase.SaveAssets();

            _db.database = new List<PropertyDBRecord>();
        }
        private static void LoadDatabase()
        {
            _db = AssetDatabase.LoadAssetAtPath<PropertiesDB>(Path.Get(Path.KEY_OUTPUT_CACHE_PROPERTIES_DB));
            if (_db == null)
                CreateDatabase();
        }
        private static string OutputCachePath
        {
            get
            {
                string s = Settings.Get(Path.Resolve(Path.KEY_OUTPUT_CACHE_PROPERTIES_PATH));

                if (!Directory.Exists(s))
                    Directory.CreateDirectory(s);

                return s;
            }
        }
        private static string NameToPath(string className, string propertyName, bool virtualTokens)
        {
            string dir = "";
            if (!virtualTokens)
                dir = OutputCachePath + "/" + className;
            else
                dir = Path.Token(Path.KEY_OUTPUT_CACHE_PROPERTIES_PATH) + "/" + className;
            if (!virtualTokens && !Directory.Exists(dir))
                Directory.CreateDirectory(dir);
            return dir + "/" + propertyName + ".asset";
        }

        public static PropertyRecord GetPropertyByMachineName(string className, string propertyName)
        {
            PropertyRecord rec = AssetDatabase.LoadAssetAtPath<PropertyRecord>(NameToPath(className, propertyName, false));
            return rec;
        }

        public static bool PropertyExists(string className, string propertyName)
        {
            PropertyRecord rec = GetPropertyByMachineName(className, propertyName);
            return (rec != null);
        }

        public static bool Create(PropertyBase property, string machineClassName, string propType, bool checkForValidity = true)
        {
            if (machineClassName == "" || (checkForValidity && !ClassManager.ClassExists(machineClassName)))
            {
                EditorUtility.DisplayDialog("Can't create the property", "Class with the name \"" + machineClassName + "\" does not exists. Property can't be created.", "OK");
                return false;
            }

            if (property.MachineName == "" || (checkForValidity && PropertyExists(machineClassName, property.MachineName)))
            {
                EditorUtility.DisplayDialog("Can't create the property", "A property with the name \"" + property.MachineName + "\" already exists inside this class.", "OK");
                return false;
            }

            PropertyRecord record = ScriptableObject.CreateInstance<PropertyRecord>();
            record.PropertySerialized = property.ToSerialized(propType);
            record.ClassMachineName = machineClassName;
            record.PropertyMachineName = property.MachineName;
            record.PropertyType = propType;
            string pathPhys = NameToPath(machineClassName, property.MachineName, false);
            AssetDatabase.CreateAsset(record, pathPhys);
            DB.database.Add(new PropertyDBRecord(machineClassName, property.MachineName));

            AssetDatabase.Refresh();
            EditorUtility.SetDirty(record);
            EditorUtility.SetDirty(DB);
            AssetDatabase.SaveAssets();

            _lastManipulatedClassName = machineClassName;
            _lastManipulatedPropertyName = property.MachineName;

            return true;
        }

        private static ExposeFilter DefaultExposeFilter = new ExposeFilter();

        public static List<KeyValuePair<PropertyDBRecord, PropertyRecord>> PropertiesOfClass(string classMachineName)
        {
            return PropertiesOfClass(classMachineName, DefaultExposeFilter);
        }

        public static List<KeyValuePair<PropertyDBRecord, PropertyRecord>> PropertiesOfClass(string classMachineName, ExposeFilter filter)
        {
            List<KeyValuePair<PropertyDBRecord, PropertyRecord>> list =
                new List<KeyValuePair<PropertyDBRecord, PropertyRecord>>();
            var r = DB.database.Where(x => x.machineClassName == classMachineName);
            foreach (var x in r)
            {
                if (filter.ExposeToInstances == ExposeFilter.Filter.DontCare && filter.ExposeToTemplates == ExposeFilter.Filter.DontCare)
                {
                    list.Add(new KeyValuePair<PropertyDBRecord, PropertyRecord>(x, GetPropertyByMachineName(x.machineClassName, x.machinePropertyName)));
                }
                else
                {
                    // We need to filter.
                    var prop = GetPropertyByMachineName(x.machineClassName, x.machinePropertyName);
                    PropertyBase b = PropertyBase.FromSerialized(prop.PropertyType, prop.PropertySerialized);
                    if (filter.MatchesFilter(b.ExposeToTemplates, b.ExposeToInstances))
                    {
                        list.Add(new KeyValuePair<PropertyDBRecord, PropertyRecord>(x, GetPropertyByMachineName(x.machineClassName, x.machinePropertyName)));
                    }
                }
            }
            return list;
        }

        public static List<PropertyDBRecord> FindPropagation(PropertyRecord parent)
        {
            List<PropertyDBRecord> olist = new List<PropertyDBRecord>();
            var list = PropertiesOfClass(parent.ClassMachineName);
            var bse = PropertyBase.FromSerialized(parent.PropertyType, parent.PropertySerialized);

            foreach (var l in list)
            {
                var x = PropertyBase.FromSerialized(l.Value.PropertyType, l.Value.PropertySerialized);
                if (x.ValuePropagation == bse.MachineName)
                {
                    olist.Add(l.Key);
                }
            }
            return olist;
        }

        public static void RemoveAll(string machineClassName)
        {
            var pr = PropertiesOfClass(machineClassName);
            foreach (var p in pr)
            {
                Remove(machineClassName, p.Key.machinePropertyName);
            }
        }

        public static bool Remove(string machineClassName, string machinePropertyName, bool confirmationDialog = true, bool fixPropagation = true)
        {
            PropertyRecord rec = GetPropertyByMachineName(machineClassName, machinePropertyName);
            ClassRecord clr = ClassManager.GetClassByMachineName(machineClassName);
            if (rec == null || clr == null)
                return false;

            PropertyBase prop = PropertyBase.FromSerialized(rec.PropertyType, rec.PropertySerialized);

            if (confirmationDialog)
            {
                if (!EditorUtility.DisplayDialog(
                    "Remove property " + prop.Name + "?",
                    "Do you wish to remove property " + prop.Name + " (" + prop.MachineName + ") from the class " + clr.className + "?",
                    "Remove", "Cancel"
                    ))
                {
                    return false;
                }
            }
            AssetDatabase.Refresh();

            if (fixPropagation)
            {
                var propags = FindPropagation(rec);
                foreach (var p in propags)
                {
                    var pp = GetPropertyByMachineName(machineClassName, p.machinePropertyName);
                    PropertyBase b = PropertyBase.FromSerialized(pp.PropertyType, pp.PropertySerialized);
                    b.ValuePropagation = "";
                    pp.PropertySerialized = b.ToSerialized(pp.PropertyType);
                    EditorUtility.SetDirty(pp);
                }
            }

            PropertyDBRecord dbrec = DB.database.Find(x => x.machineClassName == machineClassName &&
                x.machinePropertyName == machinePropertyName);
            DB.database.Remove(dbrec);
            EditorUtility.SetDirty(DB);
            AssetDatabase.DeleteAsset(NameToPath(machineClassName, machinePropertyName, false));
            ScriptableObject.DestroyImmediate(rec);

            AssetDatabase.SaveAssets();

            return true;
        }

        public static void RenameClassRutine(string oldName, string newName)
        {
            if (oldName == newName)
                return;

            var all = PropertiesOfClass(oldName);
            foreach(var a in all)
            {
                var dbrec = a.Key;
                var rec = a.Value;

                dbrec.machineClassName = newName;

                rec.ClassMachineName = newName;
                EditorUtility.SetDirty(rec);
            }
            EditorUtility.SetDirty(DB);

            AssetDatabase.SaveAssets();

            var olddir = OutputCachePath + "/" + oldName;
            var newdir = OutputCachePath + "/" + newName;
            if (Directory.Exists(olddir))
                AssetDatabase.MoveAsset(olddir, newdir);

            AssetDatabase.SaveAssets();
        }

        public static bool Edit(PropertyBase newProperty, string oldPropName, string machineClassName, string propType, bool checkForValidity = true)
        {
            var byName = GetPropertyByMachineName(machineClassName, newProperty.MachineName);
            if (newProperty.MachineName == "" || (byName != null && newProperty.MachineName != oldPropName))
            {
                EditorUtility.DisplayDialog("Can't edit the attribute", "Attribute with the name \"" + newProperty.MachineName + "\" already exists for this class, or the name is invalid.", "OK");
                return false;
            }


            PropertyDBRecord dbrec = DB.database.Find(x => x.machineClassName == machineClassName &&
                x.machinePropertyName == oldPropName);

            AssetDatabase.Refresh();

            if (oldPropName != newProperty.MachineName)
            {
                // We need to rename.
                var old = GetPropertyByMachineName(machineClassName, oldPropName);

                var propags = FindPropagation(old);
                foreach (var p in propags)
                {
                    var pp = GetPropertyByMachineName(machineClassName, p.machinePropertyName);
                    PropertyBase b = PropertyBase.FromSerialized(pp.PropertyType, pp.PropertySerialized);
                    b.ValuePropagation = newProperty.MachineName;
                    pp.PropertySerialized = b.ToSerialized(pp.PropertyType);
                    EditorUtility.SetDirty(pp);
                }

                var oldr = PropertyBase.FromSerialized(old.PropertyType, old.PropertySerialized);

                DB.database.Remove(dbrec);
                AssetDatabase.DeleteAsset(NameToPath(old.ClassMachineName, oldr.MachineName, false));
                ScriptableObject.DestroyImmediate(old);
                PropertyRecord record = ScriptableObject.CreateInstance<PropertyRecord>();
                record.PropertySerialized = newProperty.ToSerialized(propType);
                record.ClassMachineName = machineClassName;
                record.PropertyMachineName = newProperty.MachineName;
                record.PropertyType = propType;
                string pathPhys = NameToPath(machineClassName, newProperty.MachineName, false);
                AssetDatabase.CreateAsset(record, pathPhys);
                DB.database.Add(new PropertyDBRecord(machineClassName, newProperty.MachineName));
                EditorUtility.SetDirty(record);
                EditorUtility.SetDirty(DB);
            }
            else
            {
                var record = GetPropertyByMachineName(machineClassName, oldPropName);
                record.PropertySerialized = newProperty.ToSerialized(propType);
                EditorUtility.SetDirty(record);
            }

            AssetDatabase.SaveAssets();

            return true;
        }

        public static PropertyRecord ClonedProperty(PropertyRecord original)
        {
            PropertyRecord pr = ScriptableObject.CreateInstance<PropertyRecord>();
            pr.ClassMachineName = original.ClassMachineName;
            pr.PropertyMachineName = original.PropertyMachineName;
            pr.PropertyType = original.PropertyType;
            pr.PropertySerialized = original.PropertySerialized;
            return pr;
        }
        
    }

}