﻿using System.Collections;
using System.Collections.Generic;
using System.IO;
using UnityEditor;
using UnityEngine;

namespace UPP.Engine
{
    public class TemplateManager
    {
        private static TemplatesDB _db = null;
        public static TemplatesDB DB
        {
            get
            {
                if (_db == null)
                    LoadDatabase();
                return _db;
            }
        }

        private static string _lastManipulatedTemplateName = "";
        public static string LastManipulatedTemplateName { get { return _lastManipulatedTemplateName; } }

        private static void CreateDatabase()
        {
            _db = ScriptableObject.CreateInstance<TemplatesDB>();
            AssetDatabase.CreateAsset(_db, Path.Get(Path.KEY_OUTPUT_CACHE_TEMPLATES_DB));

            AssetDatabase.Refresh();
            EditorUtility.SetDirty(_db);
            AssetDatabase.SaveAssets();

            _db.database = new List<TemplateDBRecord>();
        }

        private static void LoadDatabase()
        {
            _db = AssetDatabase.LoadAssetAtPath<TemplatesDB>(Path.Get(Path.KEY_OUTPUT_CACHE_TEMPLATES_DB));
            if (_db == null)
                CreateDatabase();
        }

        private static string OutputCachePath
        {
            get
            {
                string s = Settings.Get(Path.Resolve(Path.KEY_OUTPUT_CACHE_TEMPLATES_PATH));

                if (!Directory.Exists(s))
                    Directory.CreateDirectory(s);

                return s;
            }
        }
        private static string NameToPath(string name, bool virtualTokens)
        {
            if (!virtualTokens)
                return OutputCachePath + "/" + name + ".asset";
            else
                return Path.Token(Path.KEY_OUTPUT_CACHE_TEMPLATES_PATH) + "/" + name + ".asset";
        }
        private static string NameToPathProperty(string tempName, string propName, bool virtualTokens)
        {
            if (!virtualTokens)
            {
                string s = OutputCachePath + "/" + tempName;
                if (!Directory.Exists(s))
                    Directory.CreateDirectory(s);
                return s + "/" + propName + ".asset";
            }
            else
                return Path.Token(Path.KEY_OUTPUT_CACHE_TEMPLATES_PATH) + "/" + tempName + "/" + propName + ".asset";
        }
        public static TemplateRecord GetTemplateByMachineName(string name)
        {
            TemplateRecord rec = AssetDatabase.LoadAssetAtPath<TemplateRecord>(NameToPath(name, false));
            return rec;
        }
        public static bool TemplateExists(string name)
        {
            TemplateRecord rec = GetTemplateByMachineName(name);
            return (rec != null);
        }
        public static bool TemplateOrClassExists(string name)
        {
            return TemplateExists(name) || ClassManager.ClassExists(name);
        }

        public static List<KeyValuePair<TemplateDBRecord, TemplateRecord>> FindTemplatesOfType(string parentalName)
        {
            List<KeyValuePair<TemplateDBRecord, TemplateRecord>> list =
                new List<KeyValuePair<TemplateDBRecord, TemplateRecord>>();
            foreach (var dbrec in DB.database)
            {
                if (dbrec.parentClass == parentalName)
                {
                    var rec = GetTemplateByMachineName(dbrec.machineName);
                    list.Add(new KeyValuePair<TemplateDBRecord, TemplateRecord>(dbrec, rec));
                }
            }
            return list;
        }

        public static bool Create(string templateName, string machineName, string parentalClass, string category = "", bool checkForExistence = true)
        {
            if (machineName == "" || (checkForExistence && TemplateOrClassExists(machineName)))
            {
                EditorUtility.DisplayDialog("Can't create the template", "Class or template with the name \"" + machineName + "\" already exists, or the name is invalid.", "OK");
                return false;
            }

            TemplateRecord newTemp = ScriptableObject.CreateInstance<TemplateRecord>();
            newTemp.TemplateName = templateName;
            newTemp.TemplateMachineName = machineName;
            newTemp.Category = category;
            newTemp.MachineParent = parentalClass;
            newTemp.PropertyRecords = new List<PropertyRecord>();
            string pathPhys = NameToPath(machineName, false);
            AssetDatabase.CreateAsset(newTemp, pathPhys);

            DB.database.Add(new TemplateDBRecord(machineName, parentalClass));

            AssetDatabase.Refresh();
            EditorUtility.SetDirty(newTemp);
            EditorUtility.SetDirty(DB);
            AssetDatabase.SaveAssets();
            _lastManipulatedTemplateName = machineName;
            return true;
        }

        public static bool Edit(TemplateRecord rec, string name, string machineName, string category = "")
        {
            var byName = GetTemplateByMachineName(machineName);
            var byNameCl = ClassManager.GetClassByMachineName(machineName);
            if (machineName == "" || (byName != null && byName != rec) || byNameCl != null)
            {
                EditorUtility.DisplayDialog("Can't edit the template", "Template with the name \"" + machineName + "\" already exists, or the name is invalid.", "OK");
                return false;
            }

            string oldName = rec.TemplateMachineName;

            if (oldName != machineName)
            {
                TemplateDBRecord dbrec = DB.database.Find(x => x.machineName == oldName);
                DB.database.Remove(dbrec);

                TemplateRecord newTemp = ScriptableObject.CreateInstance<TemplateRecord>();
                newTemp.TemplateName = name;
                newTemp.TemplateMachineName = machineName;
                newTemp.Category = category;
                newTemp.MachineParent = rec.MachineParent;
                newTemp.PropertyRecords = new List<PropertyRecord>(rec.PropertyRecords);
                string pathPhys = NameToPath(machineName, false);
                AssetDatabase.CreateAsset(newTemp, pathPhys);

                AssetDatabase.DeleteAsset(NameToPath(rec.TemplateMachineName, false));
                ScriptableObject.DestroyImmediate(rec);

                AssetDatabase.Refresh();

                DB.database.Add(new TemplateDBRecord(machineName, newTemp.MachineParent));
                EditorUtility.SetDirty(newTemp);
                EditorUtility.SetDirty(DB);
            }
            else
            {
                rec.TemplateName = name;
                rec.TemplateMachineName = machineName;
                rec.Category = category;
                EditorUtility.SetDirty(rec);
            }

            AssetDatabase.SaveAssets();
            _lastManipulatedTemplateName = machineName;

            return true;
        }

        public static bool Remove(string machineName, bool confirmationDialog = true)
        {
            var rec = GetTemplateByMachineName(machineName);
            if (rec == null)
                return false;

            if (confirmationDialog)
            {
                if (!EditorUtility.DisplayDialog(
                    "Remove template " + rec.TemplateName + "?",
                    "Do you wish to remove template " + rec.TemplateName + " (" + rec.TemplateMachineName + ")?",
                    "Remove", "Cancel"
                    ))
                {
                    return false;
                }
            }

            AssetDatabase.Refresh();
            TemplateDBRecord dbrec = DB.database.Find(x => x.machineName == machineName);
            DB.database.Remove(dbrec);
            EditorUtility.SetDirty(DB);
            AssetDatabase.DeleteAsset(NameToPath(machineName, false));
            ScriptableObject.DestroyImmediate(rec);

            AssetDatabase.SaveAssets();

            return true;
        }

        public static List<PropertyRecord> FindAllAtributesFor(TemplateRecord record, bool filterForTemplate = true)
        {
            var list = new List<PropertyRecord>();

            string parent = record.MachineParent;
            while (parent != "")
            {
                ClassRecord cls = ClassManager.GetClassByMachineName(parent);
                if (cls == null)
                    break;

                List<KeyValuePair<PropertyDBRecord, PropertyRecord>> collection = null;
                if (!filterForTemplate)
                    collection = PropertyManager.PropertiesOfClass(parent);
                else
                    collection = PropertyManager.PropertiesOfClass(parent, 
                        new PropertyManager.ExposeFilter(
                            PropertyManager.ExposeFilter.Filter.Yes,
                            PropertyManager.ExposeFilter.Filter.DontCare
                            ));

                foreach (var x in collection)
                {
                    list.Add(x.Value);
                }
                parent = cls.machineParent;
            }

            return list;
        }

        public static void PrepareAttributes(TemplateRecord record)
        {
            var originals = FindAllAtributesFor(record);
            var toRemove = new List<PropertyRecord>();
            var existing = new List<PropertyRecord>();

            AssetDatabase.Refresh();

            foreach (var propRec in record.PropertyRecords)
            {
                var property = PropertyBase.FromSerialized(propRec.PropertyType, propRec.PropertySerialized);
                string foundName = property.MachineName;

                if (!originals.Exists(x => x.PropertyMachineName == foundName))
                    // Found an old, no longer used attribute, remove it.
                    toRemove.Add(propRec);
                else
                {
                    // Found an attribute in use, need to check integrity.
                    var origs = originals.FindAll(x => x.PropertyMachineName == foundName);
                    PropertyBase blended = null;
                    foreach (var o in origs)
                    {
                        var bl = PropertyBase.Blend(o, propRec, false);
                        if (bl != null)
                        {
                            blended = bl;
                            break;
                        }
                    }
                    
                    propRec.PropertySerialized = blended.ToSerialized(propRec.PropertyType);
                    EditorUtility.SetDirty(propRec);
                    existing.Add(propRec);
                }
            }

            foreach (var old in toRemove)
            {
                record.PropertyRecords.Remove(old);
                AssetDatabase.DeleteAsset(NameToPathProperty(record.TemplateMachineName, old.PropertyMachineName, false));
                ScriptableObject.DestroyImmediate(old);
            }

            foreach (var original in originals)
            {
                string origName = original.PropertyMachineName;
                if (!existing.Exists(x => x.PropertyMachineName == origName))
                {
                    // Found a new attribute that needs to be created.
                    var blended = PropertyBase.Blend(original, original, true);
                    PropertyRecord propRec = PropertyManager.ClonedProperty(original);
                    propRec.PropertySerialized = blended.ToSerialized(propRec.PropertyType);
                    AssetDatabase.CreateAsset(propRec,
                        NameToPathProperty(record.TemplateMachineName, original.PropertyMachineName, false)
                        );
                    record.PropertyRecords.Add(propRec);
                    EditorUtility.SetDirty(propRec);

                }
            }
            EditorUtility.SetDirty(record);

            AssetDatabase.SaveAssets();
        }
    }
}