﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEditor;

namespace UPP.OGUI
{

    public abstract class Control : Container
    {
        private bool preparingSize = false;
        public bool PreparingSize { get {
                bool p = false;
                if (Parent != null)
                {
                    try
                    {
                        p = ((Control)Parent).preparingSize;
                    }
                    catch
                    {
                        p = false;
                    }
                }
                return preparingSize || p;
            } set { preparingSize = value; } }

        private Container oldParent = null;
        public Container Parent { get { return oldParent; } set { SetParent(value); } }
        private void SetParent(Container newParent)
        {
            if (oldParent == null)
            {
                oldParent = newParent;
                newParent.AddControl(this);
                return;
            }
            else if (!oldParent.Equals(newParent))
            {
                oldParent.RemoveControl(this);
                oldParent = newParent;
                if (newParent != null)
                    newParent.AddControl(this);
            }
        }

        public static Rect SetRectIntoDrawableArea(Rect rect, Region drawableArea)
        {
            Rect r = new Rect();
            r.x = rect.x + drawableArea.Position.X;
            r.y = rect.y + drawableArea.Position.Y;
            r.width = rect.width;
            r.height = rect.height;
            if (r.x + r.width > drawableArea.Position.X + drawableArea.Size.Width)
                r.width = (drawableArea.Position.X + drawableArea.Size.Width) - r.x;
            if (r.y + r.height > drawableArea.Position.Y + drawableArea.Size.Height)
                r.height = (drawableArea.Position.Y + drawableArea.Size.Height) - r.y;

            if (r.width < 0) r.width = 0;
            if (r.height < 0) r.height = 0;
            return r;
        }

        private Anchor oldAnchor;
        public Anchor Anchor { get { return oldAnchor; } set { SetAnchor(value); } }
        public delegate void AnchorChangedHandler(bool topOld, bool topNew, bool rightOld, bool rightNew, bool bottomOld, bool bottomNew, bool leftOld, bool leftNew);
        public event AnchorChangedHandler AnchorChanged;
        private void SetAnchor(Anchor newAnchor)
        {
            if (newAnchor.Top != oldAnchor.Top || newAnchor.Left != oldAnchor.Left || newAnchor.Right != oldAnchor.Right || newAnchor.Bottom != oldAnchor.Bottom)
            {
                Anchor back = new Anchor(oldAnchor.Top, oldAnchor.Right, oldAnchor.Bottom, oldAnchor.Left);
                oldAnchor = newAnchor;

                if (AnchorChanged != null)
                    AnchorChanged(back.Top, oldAnchor.Top, back.Right, oldAnchor.Right, back.Bottom, oldAnchor.Bottom, back.Left, oldAnchor.Left);

                newAnchor.AnchorChanged += NewAnchor_AnchorChanged;
            }
        }

        private void NewAnchor_AnchorChanged(bool topOld, bool topNew, bool rightOld, bool rightNew, bool bottomOld, bool bottomNew, bool leftOld, bool leftNew)
        {
            if (AnchorChanged != null)
                AnchorChanged(topOld, topNew, rightOld, rightNew, bottomOld, bottomNew, leftOld, leftNew);
        }

        private Size oldSize;
        public Size Size { get { return oldSize; } set { SetSize(value); } }
        private DockStyle oldDock;
        public DockStyle Dock { get { return oldDock; } set { SetDock(value); } }
        private void SetDock(DockStyle newDock)
        {
            if (oldDock == newDock) return;

            oldDock = newDock;
            if (this.Parent != null)
                this.Parent.RecomputeRenderQueue();
        }

        public Position oldPosition;
        public Position Position { get { return oldPosition; } set { SetPosition(value); } }

        public int Order { get; set; }
        public string Name { get; set; }

        public delegate void SizeChangedHandler(object sender, float oldWidth, float oldHeight, float newWidth, float newHeight);
        public event SizeChangedHandler SizeChanged;
        private void SetSize(Size newSize)
        {
            if (newSize.Width != oldSize.Width || newSize.Height != oldSize.Height)
            {
                Size back = new Size(oldSize.Width, oldSize.Height);
                oldSize = newSize;

                if (SizeChanged != null)
                    SizeChanged(this, back.Width, back.Height, newSize.Width, newSize.Height);

                newSize.SizeChanged += NewSize_SizeChanged;
                ResizeTellChildren(back.Width, back.Height, newSize.Width, newSize.Height);
            }
        }

        private void ResizeTellChildren(float oldWidth, float oldHeight, float newWidth, float newHeight)
        {
            if (preparingSize)
                return;

            List<Control> ctrls = this.GetControls();
            foreach (Control ctrl in ctrls)
            {

                if (ctrl.Dock == DockStyle.None)
                {
                    Position newPos = new Position(ctrl.Position.X, ctrl.Position.Y);
                    Size newSize = new Size(ctrl.Size.Width, ctrl.Size.Height);

                    if (ctrl.Anchor.Right && !ctrl.Anchor.Left)
                    {
                        // The control is "anchored" to the right size but not the left
                        // - that is the only situation that leads to changing the
                        // horizontal position
                        newPos.X += (newWidth - oldWidth);
                    }
                    if (ctrl.Anchor.Bottom && !ctrl.Anchor.Top)
                    {
                        // The control is "anchored" to the bottom size but not the top
                        // - that is the only situation that leads to changing the
                        // vertical position
                        newPos.Y += (newHeight - oldHeight);
                    }
                    if (ctrl.Anchor.Left && ctrl.Anchor.Right)
                    {
                        // The control is "anchored" from both horizontal sides
                        // - that is the only situation that leads to horizontal resize
                        newSize.Width += (newWidth - oldWidth);
                    }
                    if (ctrl.Anchor.Top && ctrl.Anchor.Bottom)
                    {
                        // The control is "anchored" from both vertical sides
                        // - that is the only situation that leads to vertical resize
                        newSize.Height += (newHeight - oldHeight);
                    }
                    if (!ctrl.Anchor.Left && !ctrl.Anchor.Right)
                    {
                        // Floating in the middle (no anchor) - horizontal
                        newPos.X += (newWidth - oldWidth) / 2;
                    }
                    if (!ctrl.Anchor.Left && !ctrl.Anchor.Right)
                    {
                        // Floating in the middle (no anchor) - vertical
                        newPos.Y += (newHeight - oldHeight) / 2;
                    }

                    ctrl.Size = newSize;
                    ctrl.Position = newPos;
                }
            }
        }

        public delegate void PositionChangedHandler(object sender, float oldX, float oldY, float newX, float newY);
        public event PositionChangedHandler PositionChanged;
        private void SetPosition(Position newPos)
        {
            if (newPos.X != oldPosition.X || newPos.Y != oldPosition.Y)
            {
                Position back = new Position(oldPosition.X, oldPosition.Y);
                oldPosition = newPos;

                if (PositionChanged != null)
                    PositionChanged(this, back.X, back.Y, newPos.X, newPos.Y);

                newPos.PositionChanged += NewPos_PositionChanged;
            }
        }

        private void NewPos_PositionChanged(float oldX, float oldY, float newX, float newY)
        {
            if (PositionChanged != null)
                PositionChanged(this, oldX, oldY, newX, newY);
        }

        private void NewSize_SizeChanged(float oldWidth, float oldHeight, float newWidth, float newHeight)
        {
            ResizeTellChildren(oldWidth, oldHeight, newWidth, newHeight);
            if (SizeChanged != null)
                SizeChanged(this, oldWidth, oldHeight, newWidth, newHeight);
        }

        public void SendToBack()
        {
            if (oldParent != null)
                oldParent.SendToBack(this);
        }

        public void BringToFront()
        {
            if (oldParent != null)
                oldParent.BringToFront(this);
        }

        private void Intialize()
        {
            this.oldSize = new Size(100, 100);
            this.oldPosition = new Position(0, 0);
            this.Dock = DockStyle.None;
            this.oldAnchor = new Anchor();

            this.Order = -1;
            this.Name = "";

            this.Anchor.AnchorChanged += NewAnchor_AnchorChanged;
            this.Size.SizeChanged += NewSize_SizeChanged;
            this.Position.PositionChanged += NewPos_PositionChanged;
        }

        public Control()
        {
            this.Intialize();
        }

        public Control (Control parent)
        {
            this.Intialize();
            this.oldParent = parent;
            parent.AddControl(this);
        }

        public Position AbsolutePosition { get {
                float x = this.Position.X;
                float y = this.Position.Y;
                if (this.Parent != null && this.Parent is Control)
                {
                    Position par = (this.Parent as Control).AbsolutePosition;
                    x += par.X;
                    y += par.Y;
                }
                return new Position(x, y);
            } }

        private Object _windowParent = null;
        public Object WindowParent {
            set { _windowParent = value; }
            get
            {
                if (this._windowParent != null)
                {
                    return this._windowParent;
                }
                if (this.Parent != null && this.Parent is Control)
                {
                    return (this.Parent as Control).WindowParent;
                }
                return null;
            }
        }

        public abstract string ControlName();
        public abstract void Draw(ref Region drawableArea);

        
    }

}