﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System.Linq;
using UnityEditor;

namespace UPP.OGUI
{
    public class ComboBox : Inline
    {
        private List<KeyValuePair<string, string>> captions = new List<KeyValuePair<string, string>>();
        private string emptyLabel = "";
        public string EmptyLabel { get { return emptyLabel; } set { changeCaption(value); } }

        private int _selectedIndex = -1;
        public int SelectedIndex { get { return _selectedIndex; } set { ChangeIndex(value); } }

        public string SelectedText {
            get
            {
                if (_selectedIndex > -1)
                    if (emptyLabel == "")
                        return captions.ElementAt(_selectedIndex).Value;
                    else
                        return captions.ElementAt(_selectedIndex + 1).Value;
                else
                    return "";
            }

            set { ChangeByText(value); }
        }

        public string SelectedKey
        {
            get
            {
                if (_selectedIndex > -1)
                    if (emptyLabel == "")
                        return captions.ElementAt(_selectedIndex).Key;
                    else
                        return captions.ElementAt(_selectedIndex + 1).Key;
                else
                    return "";
            }

            set { ChangeByKey(value); }
        }

        private void changeCaption(string newCaption)
        {
            if (emptyLabel == newCaption)
                return;

            if (newCaption == "")
            {
                captions.RemoveAt(0);
                if (SelectedIndex > -1)
                    SelectedIndex--;
                emptyLabel = "";
            }

            if (emptyLabel == "")
            {
                captions.Insert(0, new KeyValuePair<string, string>("__NONE__", newCaption));
                if (SelectedIndex > -1)
                    SelectedIndex++;
            }
            else
            {
                captions[0] = new KeyValuePair<string, string>("__NONE__", newCaption);
            }
            emptyLabel = newCaption;
        }

        private void Init(List<KeyValuePair<string, string>> data, string selectedKey, string emptyCaption)
        {
            Clear();
            captions = new List<KeyValuePair<string, string>>(data);
            if (selectedKey != "")
                SelectedKey = selectedKey;
            else
                SelectedIndex = -1;

            if (emptyCaption != "")
            {
                captions.Insert(0, new KeyValuePair<string, string>("__NONE__", emptyCaption));
                emptyLabel = emptyCaption;
                if (SelectedIndex > -1)
                    SelectedIndex++;
            }

            this.InheritedContent = DrawInherited;
            SetSizeFromContent();
            this.AutoSize = false;
            this.Size = new UPP.Size(150, GetCurrentStyle().CalcHeight(new GUIContent(), 150)
                + GetCurrentStyle().padding.top + GetCurrentStyle().padding.bottom);
        }

        public ComboBox()
        {
            Init(new List<KeyValuePair<string, string>>(), "", "");
        }

        List<KeyValuePair<string, string>> dictFromArr(string[] dataTexts)
        {
            List<KeyValuePair<string, string>> d = new List<KeyValuePair<string, string>>();
            int k = 0;
            foreach (string txt in dataTexts)
            {
                d.Add(new KeyValuePair<string, string>("KEY_" + k, txt));
                k++;
            }
            return d;
        }

        public ComboBox(List<KeyValuePair<string, string>> data)
        {
            Init(data, "", "");
        }

        public ComboBox(string[] dataTexts)
        {
            Init(dictFromArr(dataTexts), "", "");
        }

        public ComboBox(string emptyCaption, string[] dataTexts)
        {
            Init(dictFromArr(dataTexts), "", emptyCaption);
        }

        public ComboBox(string emptyCaption, List<KeyValuePair<string, string>> data)
        {
            Init(data, "", emptyCaption);
        }

        public ComboBox(string emptyCaption)
        {
            Init(new List<KeyValuePair<string, string>>(), "", emptyCaption);
        }

        public ComboBox(List<KeyValuePair<string, string>> data, string selectedKey)
        {
            Init(data, selectedKey, "");
        }

        public ComboBox(string[] dataTexts, string selectedKey)
        {
            Init(dictFromArr(dataTexts), selectedKey, "");
        }

        public ComboBox(string emptyCaption, string[] dataTexts, string selectedKey)
        {
            Init(dictFromArr(dataTexts), selectedKey, emptyCaption);
        }

        public ComboBox(string emptyCaption, List<KeyValuePair<string, string>> data, string selectedKey)
        {
            Init(data, selectedKey, emptyCaption);
        }

        private string[] arr()
        {
            string[] a = new string[captions.Count];
            int i = 0;
            foreach (KeyValuePair<string, string> pair in captions)
            {
                a[i] = pair.Value;
                i++;
            }
            return a;
        }

        private void DrawInherited()
        {
            int now = _selectedIndex;

            if (EmptyLabel != "")
                now++;

            int s = 0;

            if (AutoSize)
            {
                s = EditorGUILayout.Popup(now, arr(), GetCurrentStyle());
            }
            else
            {
                s = EditorGUILayout.Popup(now, arr(),
                    GUILayout.Width(this.Size.Width - GetCurrentStyle().padding.left - GetCurrentStyle().padding.right),
                    GUILayout.Height(this.Size.Height - GetCurrentStyle().padding.top - GetCurrentStyle().padding.bottom)
                    );
            }

            if (EmptyLabel != "")
                s--;

            if (s != _selectedIndex)
                SelectedIndex = s; // Also will trigger the event.
        }

        private void ChangeByText(string text)
        {
            int i = 0;
            foreach (KeyValuePair<string, string> pair in captions)
            {
                if (pair.Value == text)
                {
                    if (emptyLabel == "")
                        ChangeIndex(i);
                    else
                        ChangeIndex(i-1);
                    return;
                }
                i++;
            }
        }

        private void ChangeByKey(string text)
        {
            int i = 0;
            foreach (KeyValuePair<string, string> pair in captions)
            {
                if (pair.Key == text)
                {
                    if (emptyLabel == "")
                        ChangeIndex(i);
                    else
                        ChangeIndex(i - 1);
                    return;
                }
                i++;
            }
        }

        private void ChangeIndex(int newIndex)
        {
            if (newIndex == _selectedIndex)
                return;

            int num = captions.Count;
            if (emptyLabel != "")
                num--;

            if (newIndex == -1 || newIndex < num)
                _selectedIndex = newIndex;

            if (OnChange != null)
                OnChange(this);
        }

        public void Add(string newKey, string newCaption, bool autoSelect = false)
        {
            captions.Add(new KeyValuePair<string, string>(newKey, newCaption));
            if (autoSelect)
                SelectedKey = newKey;
        }

        public void Remove(string key)
        {
            int i = 0;
            foreach (KeyValuePair<string, string> pair in captions)
            {
                if (pair.Key == key)
                {
                    if (_selectedIndex >= i)
                        _selectedIndex--;
                    captions.RemoveAt(i);
                    break;
                }
                i++;
            }
        }

        public void Remove(int index)
        {
            if (_selectedIndex >= index)
                _selectedIndex--;
            captions.RemoveAt(index);
        }

        public void Clear()
        {
            _selectedIndex = -1;
            captions.Clear();
            if (emptyLabel != "")
            {
                captions.Add(new KeyValuePair<string, string>("__NONE__", emptyLabel));
            }
        }

        public delegate void OnChangeHandler(object sender);
        public event OnChangeHandler OnChange;

        public override GUIStyle DefaultStyle()
        {
            GUIStyle skin = new GUIStyle(EditorStyles.popup);
            skin.padding = new RectOffset(5, 5, 2, 2);
            return new GUIStyle(skin);
        }

        public override GUIContent GetContent()
        {
            if (_selectedIndex > -1)
                return new GUIContent(captions.ElementAt(_selectedIndex).Value);
            else
                return new GUIContent(emptyLabel);
        }

        
    }
}
