﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEditor;

namespace UPP.OGUI
{
    public class DiagramEntity : Panel {

        private string _title = "";
        public string Title { get {return _title; } set { _title = value; } }
        private IList _content;
        private IList _buttons;
        private int windowId = -1;

        private bool _selected = false;
        public bool Selected { get { return _selected; } set { SetSelected(value); } }

        private GUIStyle titleStyle = null;

        public void SetContent(IList content)
        {
            _content = content;
        }

        public void SetButtons(IList buttons)
        {
            _buttons = buttons;
        }

        private void SetSelected(bool val)
        {
            _selected = val;
            if (val)
                this.BackgroundColor = ColorSchemes.SelectionColor;
            else
                this.BackgroundColor = ColorSchemes.Transparent;
        }

        public string MachineName { get; set; }

        private Window wParent;

        private DiagramEntity _entityParent = null;
        public DiagramEntity EntityParent { get { return _entityParent; } set { SetEntityParent(value); } }
        private List<DiagramEntity> _entityChildren = new List<DiagramEntity>();

        public List<DiagramEntity> EntityChildren { get { return _entityChildren; } }

        private void SetEntityParent(DiagramEntity newParent)
        {
            if (_entityParent == newParent)
                return;

            if (_entityParent != null)
                _entityParent.RemoveEntityChild(this);
            _entityParent = newParent;
            if (_entityParent != null)
                _entityParent.AddEntityChild(this);
        }
        public void RemoveEntityChild(DiagramEntity child)
        {
            if (_entityChildren.Contains(child))
            {
                child._entityParent = null;
                _entityChildren.Remove(child);
            }
        }

        public void AddEntityChild(DiagramEntity child)
        {
            if (!_entityChildren.Contains(child))
            {
                child.EntityParent = null;
                _entityChildren.Add(child);
                child._entityParent = this;
            }
        }

        private void Init(Window winParent, string title, IList<string> content, IList<string> buttons)
        {
            wParent = winParent;
            wParent.RegisterDiagramEntity(this);
            _title = title;
            this.InheritedContent = DrawInherited;
            windowId = IDGenerator.Get();
            this.Size = new Size(175, 85);

            this.PositionChanged += DiagramEntity_PositionChanged;
            titleStyle = GUI.skin.window;
            titleStyle.fontStyle = FontStyle.Bold;
        }

        bool _wasRepositioned = false;

        private void DiagramEntity_PositionChanged(object sender, float oldX, float oldY, float newX, float newY)
        {
            if (wasClicked)
                _wasRepositioned = true;
        }

        public DiagramEntity(Window winParent, string title, IList<string> content, IList<string> buttons)
        {
            Init(winParent, title, content, buttons);
        }
        public DiagramEntity(Window winParent, string title)
        {
            Init(winParent, title, new List<string>(), new List<string>());
        }

        public Rect EntityRect()
        {
            return new Rect(Position.X + 3, Position.Y + 3, Size.Width - 7, Size.Height - 6);
        }

        private void DrawInherited()
        {

            if (wParent)
            {
                Rect oldRec = EntityRect();
                Rect newRec = GUI.Window(windowId, oldRec, DrawNodeWindow, _title, titleStyle);

                if (oldRec.x != oldRec.x || oldRec.y != newRec.y)
                {
                    Position = new Position(
                        Position.X + (newRec.x - oldRec.x),
                        Position.Y + (newRec.y - oldRec.y));
                }
            }
        }

        public delegate void OnClickHandler(object sender, int windowId);
        public event OnClickHandler OnClick;

        public delegate void OnMouseUpHandler(object sender, int windowId, bool wasRepositioned);
        public event OnMouseUpHandler OnMouseUp;

        public delegate void OnButtonClickHandler(object sender, int windowId, int buttonIndex, string buttonText);
        public event OnButtonClickHandler OnButtonClick;

        bool wasClicked = false;
        private void DrawNodeWindow(int id)
        {
            Event e = Event.current;
            Rect r = new Rect(0, 0, Size.Width, Size.Height);
            if (e.type == EventType.MouseDown && e.button == 0 && r.Contains(e.mousePosition))
            {
                _wasRepositioned = false;
                wasClicked = true;
                if (OnClick != null)
                    OnClick(this, id);
            }

            if (e.type == EventType.MouseUp && e.button == 0 && wasClicked)
            {
                wasClicked = false;
                if (OnMouseUp != null)
                    OnMouseUp(this, id, _wasRepositioned);
                _wasRepositioned = false;
            }

            int y = 20;
            const int lineHeight = 16;
            foreach (string s in _content)
            {
                string[] spl = s.Split(':');
                if (spl.Length != 2)
                    GUI.Label(new Rect(3, y, Size.Width - 6, lineHeight), s);
                else
                {
                    GUI.Label(new Rect(3, y, 60, lineHeight), spl[0] + ":");
                    GUI.Label(new Rect(61, y, Size.Width - 66, lineHeight), spl[1]);
                }
                y += lineHeight;
            }

            float x = 2;
            float per = (Size.Width - 10) / _buttons.Count;
            int index = 0;
            foreach (string b in _buttons)
            {
                bool click = GUI.Button(new Rect(x, Size.Height - 24, per, lineHeight), b);
                x += per;

                if (click && OnButtonClick != null)
                    OnButtonClick(this, windowId, index, b);
                index++;
            }

            GUI.DragWindow();
        }

        public void DrawCurvesToChildrenRecursive()
        {
            foreach (var c in _entityChildren)
            {
                DrawCurveToChild(c);
                c.DrawCurvesToChildrenRecursive();
            }
        }

        void DrawCurveToChild(DiagramEntity child)
        {
            Rect start = EntityRect();
            Rect end = child.EntityRect();
            Vector3 startPos = new Vector3(start.x + start.width / 2, start.y + start.height, 0);
            Vector3 endPos = new Vector3(end.x + end.width / 2, end.y, 0);
            Vector3 startTan = startPos + Vector3.up * 30;
            Vector3 endTan = endPos + Vector3.down * 30;
            Color shadowCol = new Color(0, 0, 0, 0.06f);
            for (int i = 0; i < 3; i++) // Draw a shadow
                Handles.DrawBezier(startPos, endPos, startTan, endTan, shadowCol, null, (i + 1) * 5);
            Handles.DrawBezier(startPos, endPos, startTan, endTan, Color.black, null, 1);
        }
    }
}