﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEditor;

namespace UPP.OGUI
{

    public class Panel : Control
    {
        public override string ControlName() { return "Panel"; }

        private bool _visible = true;
        public bool Visible { get { return _visible; } set { _visible = value; } }

        private bool _enabled = true;
        public bool Enabled { get { return _enabled; } set { _enabled = value; } }

        private bool _isEntityDiagramCanvas = false;
        public bool IsEntityDiagramCanvas { get { return _isEntityDiagramCanvas; } set { _isEntityDiagramCanvas = value; } }

        public override Region GetDrawableArea()
        {
            if (this.SizeMode == SizeModeType.Clip)
                return new Region(0, 0, this.Size.Width, this.Size.Height);
            else
            {
                Vector2 v = GetInnerSize();
                return new Region(0, 0, v.x, v.y);
            }
        }

        private GUIStyle style;

        private MouseCursor _cursor = MouseCursor.Arrow;
        public MouseCursor Cursor { get { return _cursor; } set { _cursor = value; } }

        private Color oldBackgroundColor;
        public Color BackgroundColor { get { return oldBackgroundColor; } set { SetBackgroundColor(value); } }
        private void SetBackgroundColor(Color newColor)
        {
            if (!newColor.Equals(oldBackgroundColor))
            {
                oldBackgroundColor = newColor;
                SetStyle();
            }
        }

        private Color oldBorderColor;
        public Color BorderColor { get { return oldBorderColor; } set { SetBorderColor(value); } }
        private void SetBorderColor(Color newColor)
        {
            if (!newColor.Equals(oldBorderColor))
            {
                oldBorderColor = newColor;
                SetStyle();
            }
        }

        public enum SizeModeType { Clip, ScrollBars }
        private SizeModeType _sizeMode = SizeModeType.Clip;
        public SizeModeType SizeMode { get { return _sizeMode; } set {
                _sizeMode = value;
                UnityEngine.Object par = this.WindowParent;
                if (par != null && par is EditorWindow)
                    (par as EditorWindow).Repaint();
        } }

        private void Initialize()
        {
            oldBackgroundColor = new Color(0, 0, 0, 0);
            oldBorderColor = new Color(0, 0, 0, 0);

            style = new GUIStyle(GUIStyle.none);
            style.margin = new RectOffset(0, 0, 0, 0);
            style.padding = new RectOffset(0, 0, 0, 0);

            this.SizeChanged += SizeChangedRoutine;

            SetStyle();
        }

        private void SetStyle()
        {
            if (this.Size.Width <= 0 || this.Size.Height <= 0)
                return;

            Texture2D texture = new Texture2D((int)this.Size.Width, (int)this.Size.Height);

            ColorSchemes.FillTexture(texture, this.BackgroundColor);

            if (this.oldBorderColor.a > 0)
            {
                if (Dock == DockStyle.None || Dock == DockStyle.Top)
                    texture.DrawLineHorizontal(0, (int)Size.Width, 0, oldBorderColor);
                if (Dock == DockStyle.None || Dock == DockStyle.Bottom)
                    texture.DrawLineHorizontal(0, (int)Size.Width, (int)Size.Height - 1, oldBorderColor);
                if (Dock == DockStyle.None || Dock == DockStyle.Right)
                    texture.DrawLineVertical(0, 0, (int)Size.Height, oldBorderColor);
                if (Dock == DockStyle.None || Dock == DockStyle.Left)
                texture.DrawLineVertical((int)Size.Width - 1, 0, (int)Size.Height, oldBorderColor);

                texture.Apply();
            }

            style.normal.background = texture;
        }

        private void SizeChangedRoutine(object sender, float ow, float oh, float nw, float nh)
        {
            SetStyle();
        }

        public Panel() : base() { Initialize(); }
        public Panel(Control parent) : base(parent) { Initialize(); }

        protected Vector2 ScrollPos = new Vector2();
        

        public Vector2 GetInnerSize()
        {
            float sx = 0, sy = 0;
            foreach (Control c in this.GetControls())
            {
                float cx = c.Position.X + c.Size.Width;
                float cy = c.Position.Y + c.Size.Height;
                if (cx > sx) sx = cx;
                if (cy > sy) sy = cy;
            }

            return new Vector2(sx, sy);
        }

        public override void Draw(ref Region drawableArea)
        {
            if (!Visible) return;

            if (!Enabled || (Parent != null && !(Parent as Panel).Enabled))
                GUI.enabled = false;

            Rect r = new Rect();
            if (this.Dock == DockStyle.None)
            {
                r = new Rect(this.Position.X, this.Position.Y, this.Size.Width, this.Size.Height);
                r = SetRectIntoDrawableArea(r, drawableArea);
            }

            if (this.Dock == DockStyle.Fill)
            {
                r = new Rect(drawableArea.Position.X, drawableArea.Position.Y, drawableArea.Size.Width, drawableArea.Size.Height);
                this.Size.Width = r.width;
                this.Size.Height = r.height;
                this.Position.X = drawableArea.Position.X;
                this.Position.Y = drawableArea.Position.Y;
            }

            if (this.Dock == DockStyle.Top)
            {
                r = new Rect(drawableArea.Position.X, drawableArea.Position.Y, drawableArea.Size.Width, this.Size.Height);
                this.Size.Width = r.width;
                this.Size.Height = r.height;
                this.Position.X = drawableArea.Position.X;
                this.Position.Y = drawableArea.Position.Y;
                drawableArea.Position.Y += this.Size.Height;
                drawableArea.Size.Height -= this.Size.Height;

            }

            if (this.Dock == DockStyle.Bottom)
            {
                r = new Rect(drawableArea.Position.X, drawableArea.Position.Y + drawableArea.Size.Height - this.Size.Height, drawableArea.Size.Width, this.Size.Height);
                this.Size.Width = r.width;
                this.Size.Height = r.height;
                this.Position.X = drawableArea.Position.X;
                this.Position.Y = drawableArea.Position.Y + drawableArea.Size.Height - this.Size.Height;
                drawableArea.Size.Height -= this.Size.Height;
            }

            if (this.Dock == DockStyle.Left)
            {
                r = new Rect(drawableArea.Position.X, drawableArea.Position.Y, this.Size.Width, drawableArea.Size.Height);
                this.Size.Width = r.width;
                this.Size.Height = r.height;
                this.Position.X = drawableArea.Position.X;
                this.Position.Y = drawableArea.Position.Y;
                drawableArea.Position.X += this.Size.Width;
                drawableArea.Size.Width -= this.Size.Width;
            }

            if (this.Dock == DockStyle.Right)
            {
                r = new Rect(drawableArea.Position.X + drawableArea.Size.Width - this.Size.Width, drawableArea.Position.Y, this.Size.Width, drawableArea.Size.Height);
                this.Size.Width = r.width;
                this.Size.Height = r.height;
                this.Position.X = drawableArea.Position.X + drawableArea.Size.Width - this.Size.Width;
                this.Position.Y = drawableArea.Position.Y;
                drawableArea.Size.Width -= this.Size.Width;
            }

            if (this.Size.Width <= 0 || this.Size.Height <= 0)
                return;

            GUILayout.BeginArea(r, style);

            if (this.SizeMode == SizeModeType.ScrollBars)
            {
                ScrollPos = GUILayout.BeginScrollView(ScrollPos);

                Vector2 v = GetInnerSize();
                GUILayout.Label("", GUILayout.Width(v.x), GUILayout.Height(v.y));
            }

            if (_cursor != MouseCursor.Arrow)
            {
                Rect rc = new Rect(0, 0, r.width, r.height);
                EditorGUIUtility.AddCursorRect(rc, _cursor);
            }

            // Diagram Entities need to be handled through some custom code:
            if (IsEntityDiagramCanvas)
            {
                (WindowParent as Window).BeginWindows();
                var cc = GetControls();
                foreach (Control c in cc)
                {
                    if (c.GetType() == typeof(DiagramEntity))
                    {
                        var d = c as DiagramEntity;
                        if (d.EntityParent == null)
                            d.DrawCurvesToChildrenRecursive();
                    }
                }
            }

            base.DrawChildren();
            DrawInherited();

            if (IsEntityDiagramCanvas)
                (WindowParent as Window).EndWindows();

            if (this.SizeMode == SizeModeType.ScrollBars)
            {
                GUILayout.EndScrollView();
            }
            GUILayout.EndArea();

            GUI.enabled = true;
        }

        private void DrawInherited() { if (InheritedContent != null) InheritedContent(); }

        protected delegate void InheritedContentT();
        protected InheritedContentT InheritedContent = null;

        public override int GetRenderPriority()
        {
            if (this.Dock == DockStyle.None)
                return 300;
            if (this.Dock == DockStyle.Fill)
                return 200;
            return 100;
        }
    }

}