﻿using System.Collections;
using System.Collections.Generic;
using UnityEditor;
using UnityEngine;
using UPP.OGUI;

public class SplitContainer : Panel {

	public enum ModeType
    {
        Horizontal,
        Vertical
    }

    public enum SizeType
    {
        Percentage,
        Absolute
    }

    public enum PanelWho
    {
        First,
        Second
    }

    public enum ResizeReactionType
    {
        Symetrical,
        LockFirst,
        LockSecond
    }

    private ModeType _mode = ModeType.Horizontal;
    public ModeType Mode { get { return _mode; } set { SetMode(value); } }

    private ResizeReactionType _resizeReaction = ResizeReactionType.Symetrical;
    public ResizeReactionType ResizeReaction { get { return _resizeReaction; } set { _resizeReaction = value; } }

    private float _size1 = 100;
    private float _size2 = 100;

    private float _grabberWidth = 5;

    private Panel panel1;
    private Panel panel2;
    private Panel grabberGr;
    private MousePanel grabber;

    private Vector2 buttonPosClick = new Vector2();
    private float sizeOnDrag = 0;

    public Panel GetPanel(PanelWho panel)
    {
        if (panel == PanelWho.First)
            return panel1;
        else
            return panel2;
    }

    public void SetGrabberWidth(float width)
    {
        float f = (width - _grabberWidth) / 2;
        _grabberWidth = width;
        _size1 -= f;
        _size2 -= f;
        RecomputePanels(true);
    }

    private void SetMode(ModeType _newMode)
    {
        if (_newMode != _mode)
        {
            float c = GetCurrentPercentage(PanelWho.First);
            _mode = _newMode;
            SetSize(c, SizeType.Percentage, PanelWho.First);
            ActualizeGrabberCursor();
        }
    }

    private void ActualizeGrabberCursor()
    {
        if (_mode == ModeType.Horizontal)
            grabberGr.Cursor = UnityEditor.MouseCursor.ResizeHorizontal;
        else
            grabberGr.Cursor = UnityEditor.MouseCursor.ResizeVertical;
    }

    public float GetCurrentPercentage(PanelWho panel)
    {
        if (panel == PanelWho.First)
            return _size1 * 100f / (_size1 + _size2);
        else
            return _size2 * 100f / (_size1 + _size2);
    }

    public void AddToPanel(Control control, PanelWho panel)
    {
        if (panel == PanelWho.First)
            panel1.AddControl(control);
        else
            panel2.AddControl(control);
    }

    private void Init()
    {
        this.SizeChanged += SplitContainer_SizeChanged;

        panel1 = new Panel(this);
        panel2 = new Panel(this);
        grabberGr = new Panel(this);
        grabber = new MousePanel(this);

        grabberGr.BackgroundColor = UPP.ColorSchemes.BorderColor;
        grabber.OnMouseDown += Grabber_OnMouseDown;
        grabber.OnMouseUp += Grabber_OnMouseUp;
        grabber.OnMouseDrag += Grabber_OnMouseDrag;

        ActualizeGrabberCursor();
    }

    private void Grabber_OnMouseDrag(Event e)
    {
        float f;
        if (_mode == ModeType.Horizontal)
            f = e.mousePosition.x - buttonPosClick.x;
        else
            f = e.mousePosition.y - buttonPosClick.y;
        _size1 = sizeOnDrag + f;
        _size2 = GetRestFrom(_size1);
        RecomputePanels(false);

        Object window = this.WindowParent;
        if (window != null && window is EditorWindow)
        {
            buttonPosClick = e.mousePosition;
            sizeOnDrag = _size1;
            (window as EditorWindow).Repaint();
        }
    }

    private void Grabber_OnMouseUp(Event e)
    {
        float f;
        if (_mode == ModeType.Horizontal)
            f = e.mousePosition.x - buttonPosClick.x;
        else
            f = e.mousePosition.y - buttonPosClick.y;
        _size1 = sizeOnDrag + f;
        _size2 = GetRestFrom(_size1);
        RecomputePanels(true);

        Object window = this.WindowParent;
        if (window != null && window is EditorWindow)
        {
            buttonPosClick = e.mousePosition;
            sizeOnDrag = _size1;
            (window as EditorWindow).Repaint();
        }
    }

    private void Grabber_OnMouseDown(MousePanel.MouseButton button, Event e)
    {
        buttonPosClick = e.mousePosition;
        sizeOnDrag = _size1;
    }

    private void SplitContainer_SizeChanged(object sender, float oldWidth, float oldHeight, float newWidth, float newHeight)
    {
        float me = ThisSize() - _grabberWidth;

        switch (_resizeReaction)
        {
            case ResizeReactionType.LockFirst:
                _size2 = me - _size1;
                break;
            case ResizeReactionType.LockSecond:
                _size1 = me - _size2;
                break;
            default:
                float p1 = GetCurrentPercentage(PanelWho.First);
                float p2 = GetCurrentPercentage(PanelWho.Second);
                _size1 = me * p1 / 100f;
                _size2 = me * p2 / 100f;
                break;
        }

        RecomputePanels(true);
    }

    private void RecomputePanels(bool moveGrabber)
    {
        panel1.Position = new UPP.Position(0, 0);
        if (_mode == ModeType.Horizontal)
        {
            panel1.Size.Height = this.Size.Height;
            panel2.Size.Height = this.Size.Height;
            grabberGr.Size.Height = this.Size.Height;

            panel1.Size.Width = _size1;
            panel2.Size.Width = _size2;
            grabberGr.Size.Width = _grabberWidth;

            grabberGr.Position = new UPP.Position(panel1.Size.Width, 0);
            panel2.Position = new UPP.Position(grabberGr.Position.X + grabberGr.Size.Width, 0);
        }
        else
        {
            panel1.Size.Width = this.Size.Width;
            panel2.Size.Width = this.Size.Width;
            grabberGr.Size.Width = this.Size.Width;

            panel1.Size.Height = _size1;
            panel2.Size.Height = _size2;
            grabberGr.Size.Height = _grabberWidth;

            grabberGr.Position = new UPP.Position(0, panel1.Size.Height);
            panel2.Position = new UPP.Position(0, grabberGr.Position.Y + grabberGr.Size.Height);
        }

        if (moveGrabber)
        {
            grabber.Size = new UPP.Size(grabberGr.Size.Width, grabberGr.Size.Height);
            grabber.Position = new UPP.Position(grabberGr.Position.X, grabberGr.Position.Y);
        }
    }


    public SplitContainer()
    {
        _mode = ModeType.Horizontal;
        Init();
        SetSize(50, SizeType.Percentage, PanelWho.First);
    }

    public SplitContainer(ModeType mode)
    {
        _mode = mode;
        Init();
        SetSize(50, SizeType.Percentage, PanelWho.First);
    }

    public SplitContainer(float size, SizeType type, PanelWho panel)
    {
        _mode = ModeType.Horizontal;
        Init();
        SetSize(size, type, panel);
    }

    public SplitContainer(float size, SizeType type, PanelWho panel, ModeType mode)
    {
        _mode = mode;
        Init();
        SetSize(size, type, panel);
    }

    public SplitContainer(ModeType mode, float size, SizeType type, PanelWho panel)
    {
        _mode = mode;
        Init();
        SetSize(size, type, panel);
    }

    private float ThisSize()
    {
        if (this._mode == ModeType.Horizontal)
            return this.Size.Width;
        else
            return this.Size.Height;
    }

    private float GetRestFrom(float size)
    {
        return ThisSize() - _grabberWidth - size;
    }

    public void SetSize(float size, SizeType type, PanelWho panel)
    {
        float s1 = size;
        if (type == SizeType.Percentage)
        {
            s1 = ((ThisSize() - _grabberWidth) * size / 100);
        }

        if (panel == PanelWho.First)
        {
            _size1 = s1;
            _size2 = GetRestFrom(s1);
        }
        else
        {
            _size2 = s1;
            _size1 = GetRestFrom(s1);
        }
        RecomputePanels(true);
    }
}
