﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

namespace UPP.OGUI
{
    public class TreeView : Panel
    {
        public class TreeViewItem
        {
            private List<TreeViewItem> children = new List<TreeViewItem>();
            private TreeViewItem _parent = null;
            public TreeViewItem Parent { get { return _parent; } set { SetParent(value); } }
            public string MachineName { get; set; }
            public string DisplayName { get; set; }
            public int internalTreeId = 0;

            public TreeViewItem(string machineName)
            {
                MachineName = machineName;
            }

            public TreeViewItem(string machineName, string displayName)
            {
                MachineName = machineName;
                DisplayName = displayName;
            }

            public List<TreeViewItem> GetChildren()
            {
                return children;
            }

            public void RemoveChild(TreeViewItem child)
            {
                if (children.Contains(child))
                {
                    child._parent = null;
                    children.Remove(child);
                }
            }

            public void AddChild(TreeViewItem child)
            {
                if (!children.Contains(child))
                {
                    child.Parent = null;
                    children.Add(child);
                    child._parent = this;
                }
            }

            private void SetParent(TreeViewItem newParent)
            {
                if (_parent == newParent)
                    return;

                if (_parent != null)
                    _parent.RemoveChild(this);
                _parent = newParent;
                if (_parent != null)
                    _parent.AddChild(this);
            }
        }

        private class TView : UnityEditor.IMGUI.Controls.TreeView
        {
            public List<TreeViewItem> itemsRoot = new List<TreeViewItem>();
            private int idCounter = 0;

            public TView(UnityEditor.IMGUI.Controls.TreeViewState state) : base(state)
            {
                Reload();
            }

            private UnityEditor.IMGUI.Controls.TreeViewItem constructItem(TreeViewItem item, int depth)
            {
                var it = new UnityEditor.IMGUI.Controls.TreeViewItem(idCounter, depth);
                item.internalTreeId = idCounter;
                idCounter++;
                it.displayName = item.DisplayName;
                var chilren = item.GetChildren();

                foreach (TreeViewItem i in chilren)
                {
                    var c = constructItem(i, depth + 1);
                    it.AddChild(c);
                }

                return it;
            }

            protected override void DoubleClickedItem(int id)
            {
                TreeViewItem i = FindById(id);
                if (i == null)
                    return;
                if (OnDoubleClick != null)
                    OnDoubleClick(this, i.MachineName, i.DisplayName, i.internalTreeId);
            }

            protected override UnityEditor.IMGUI.Controls.TreeViewItem BuildRoot()
            {
                var root = new UnityEditor.IMGUI.Controls.TreeViewItem(-1, -1, "Root");
                idCounter = 0;

                foreach (TreeViewItem i in itemsRoot)
                {
                    var it = constructItem(i, 0);
                    root.AddChild(it);
                }

                if (itemsRoot.Count == 0)
                    root.AddChild(new UnityEditor.IMGUI.Controls.TreeViewItem(0, 0, "(empty)"));

                return root;
            }

            protected override bool CanMultiSelect(UnityEditor.IMGUI.Controls.TreeViewItem item)
            {
                return false;
            }

            public void SelectByMachineName(string key, List<TreeViewItem> list = null)
            {
                if (list == null)
                    list = itemsRoot;
                foreach (TreeViewItem i in list)
                {
                    if (i.MachineName == key)
                    {
                        SetSelection(new List<int>(new int[] { i.internalTreeId }));
                        return;
                    }
                    var c = i.GetChildren();
                    if (c != null)
                    {
                        SelectByMachineName(key, c);
                    }
                }
            }

            public TreeViewItem FindById(int internalId, List<TreeViewItem> list = null)
            {
                if (list == null)
                    list = itemsRoot;
                foreach (TreeViewItem i in list)
                {
                    if (i.internalTreeId == internalId)
                        return i;
                    var c = i.GetChildren();
                    if (c != null)
                    {
                        var x = FindById(internalId, c);
                        if (x != null)
                            return x;
                    }
                }

                return null;
            }

            public TreeViewItem FindByMachineName(string name, List<TreeViewItem> list = null)
            {
                if (list == null)
                    list = itemsRoot;
                foreach (TreeViewItem i in list)
                {
                    if (i.MachineName == name)
                        return i;
                    var c = i.GetChildren();
                    if (c != null)
                    {
                        var x = FindByMachineName(name, c);
                        if (x != null)
                            return x;
                    }
                }

                return null;
            }

            protected override void SelectionChanged(IList<int> selectedIds)
            {
                if (selectedIds.Count <= 0)
                    return;
                
                int id = selectedIds[0];
                TreeViewItem i = FindById(id);
                if (i == null)
                    return;
                if (OnSelectionChange != null)
                    OnSelectionChange(this, i.MachineName, i.DisplayName, i.internalTreeId);
            }

            public delegate void OnSelectionChangeHandler(object sender, string machineName, string displayName, int internalId);
            public event OnSelectionChangeHandler OnSelectionChange = null;
            public delegate void OnDoubleClickHandler(object sender, string machineName, string displayName, int internalId);
            public event OnDoubleClickHandler OnDoubleClick = null;
        }

        TView tView = null;

        public delegate void OnSelectionChangeHandler(object sender, string machineName, string displayName, int internalId);
        public event OnSelectionChangeHandler OnSelectionChange = null;
        public delegate void OnDoubleClickHandler(object sender, string machineName, string displayName, int internalId);
        public event OnDoubleClickHandler OnDoubleClick = null;

        public TreeView()
        {
            tView = new TView(new UnityEditor.IMGUI.Controls.TreeViewState());
            this.InheritedContent = DrawInherited;
            tView.OnSelectionChange += TView_OnSelectionChange;
            tView.OnDoubleClick += TView_OnDoubleClick;
        }

        TreeViewItem _lastSelected = null;
        public TreeViewItem SelectedItem { get { return _lastSelected; } }

        private void TView_OnDoubleClick(object sender, string machineName, string displayName, int internalId)
        {
            if (this.OnDoubleClick != null)
                OnDoubleClick(this, machineName, displayName, internalId);
        }

        private void TView_OnSelectionChange(object sender, string machineName, string displayName, int internalId)
        {
            _lastSelected = tView.FindById(internalId);
            if (this.OnSelectionChange != null)
                OnSelectionChange(this, machineName, displayName, internalId);
        }

        public void Refresh()
        {
            tView.Reload();
            tView.SetSelection(new List<int>());
            if (this.OnSelectionChange != null)
                OnSelectionChange(this, "", "", -2);
        }

        public void CollapseAll() { tView.CollapseAll(); }
        public void ExpandAll() { tView.ExpandAll(); }
        public void SetExpanded(int internalId, bool expand)
        {
            tView.SetExpanded(internalId, expand);
        }
        public void SetExpanded(string machineName, bool expand)
        {
            tView.SetExpanded(tView.FindByMachineName(machineName).internalTreeId, expand);
        }

        public void SelectByMychineName(string key)
        {
            tView.SelectByMachineName(key);
            _lastSelected = tView.FindByMachineName(key);
            if (this.OnSelectionChange != null)
                OnSelectionChange(this, _lastSelected.MachineName, _lastSelected.DisplayName, _lastSelected.internalTreeId);
        }
        public void SelectByInternalId(int id)
        {
            tView.SetSelection(new List<int>(new int[] { id }));
            _lastSelected = tView.FindById(id);
            if (this.OnSelectionChange != null)
                OnSelectionChange(this, _lastSelected.MachineName, _lastSelected.DisplayName, _lastSelected.internalTreeId);
        }

        public void SetItems(List<TreeViewItem> rootItems)
        {
            tView.itemsRoot = rootItems;
            Refresh();
        }

        public void RemoveItem(TreeViewItem item)
        {
            if (tView.itemsRoot.Contains(item))
                tView.itemsRoot.Remove(item);
            item.Parent = null;
        }

        public void Clear()
        {
            tView.itemsRoot.Clear();
        }

        public void AddItem(TreeViewItem item, TreeViewItem parent=null)
        {
            item.Parent = parent;

            if (parent == null)
            {
                tView.itemsRoot.Add(item);
            }
        }

        private void DrawInherited()
        {
            Rect r = new Rect(0, 0, this.Size.Width, this.Size.Height);
            tView.OnGUI(r);
        }

    }
}

